/*
   A rotating cube.
   Glide 2.4, Watcom 11.0 DOS, v.2
   Matrix code based upon a DrDobbs Journal example (J.Spiller, Jul94.)

   March 99: updated for Banshee, because printf/puts
   messes up the Banshee graphics.

   Bugs:
   - When the cube is too close, it slows down to a near-stall

-- Benny <benny_land@geocities.com>
*/
#include <dos.h>
#include <string.h>
#include <stdlib.h>
#include <malloc.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>
#include <assert.h>

#include <glide.h>

#define BANSHEE

#define YES	1
#define NO	0

#define SNAPPER (float)(3L << 18)
#define SNAP(vertex) \
	(vertex).x += SNAPPER; (vertex).x -= SNAPPER; \
	(vertex).y += SNAPPER; (vertex).y -= SNAPPER;

GrHwConfiguration hwconfig;

short           GlideIsRunning = NO;

short           inside = NO,
		iterated = YES,
		twosided = NO,
		allsides = YES,
		twoscreens = YES;

float           red, green, blue;

long            size_x = 640, size_y = 480;
float           side1 = 255, side2 = 255;
float           midx, midy;

long            grWireColor;

GrVertex        vtx1, vtx2, vtx3, vtx4;

class           MAT_COLUMN {
    float           Column[4];
		    public:
    float          &operator[] (int Index) {
	return Column[Index];
    }
};

class           MAT3D {
    MAT_COLUMN      Matrix[4];
		    public:
		    MAT_COLUMN & operator[] (int Index) {
	return Matrix[Index];
    } friend MAT3D operator *(MAT3D & A, MAT3D & B);
};

MAT3D operator *(MAT3D & A, MAT3D & B) {
    int             I, J, K;
    MAT3D           Temp;

    for (I = 0; I < 4; I++)
	for (J = 0; J < 4; J++) {
	    float           Sum = 0;
	    for (K = 0; K < 4; K++)
		Sum += A[I][K] * B[K][J];
	    Temp[I][J] = Sum;
	}
    return Temp;
}

MAT3D           Identity()
{
    int             X, Y;
    MAT3D           Temp;
    for (Y = 0; Y < 4; Y++)
	for (X = 0; X < 4; X++)
	    if (X == Y)
		Temp[Y][X] = 1;
	    else
		Temp[Y][X] = 0;
    return Temp;
}

MAT3D           RotateX(float Angle)
{
    MAT3D           Temp = Identity();
    Temp[1][1] = Temp[2][2] = cos(Angle);
    Temp[2][1] = -(Temp[1][2] = sin(Angle));
    return Temp;
}

MAT3D           RotateY(float Angle)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = Temp[2][2] = cos(Angle);
    Temp[0][2] = -(Temp[2][0] = sin(Angle));
    return Temp;
}

MAT3D           RotateZ(float Angle)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = Temp[1][1] = cos(Angle);
    Temp[1][0] = -(Temp[0][1] = sin(Angle));
    return Temp;
}

MAT3D           RotateXYZ(float AngleX, float AngleY, float AngleZ)
{
    MAT3D
    Temp3 = RotateX(AngleX),
    Temp2 = RotateY(AngleY),
    Temp1 = RotateZ(AngleZ);

    Temp2 = Temp2 * Temp3;
    return Temp1 * Temp2;
}

MAT3D           Scale(float Scale)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = Temp[1][1] = Temp[2][2] = Scale;
    return Temp;
}

MAT3D           ScaleXYZ(float SX, float SY, float SZ)
{
    MAT3D           Temp = Identity();
    Temp[0][0] = SX;
    Temp[1][1] = SY;
    Temp[2][2] = SZ;
    return Temp;
}

MAT3D           Translate(float X, float Y, float Z)
{
    MAT3D           Temp = Identity();
    Temp[0][3] = X;
    Temp[1][3] = Y;
    Temp[2][3] = Z;
    return Temp;
}

void            Translate_Point(MAT3D & M, float &X, float &Y, float &Z)
{
    float           TX, TY, TZ;
    TX = X * M[0][0] + Y * M[0][1] + Z * M[0][2] + M[0][3];
    TY = X * M[1][0] + Y * M[1][1] + Z * M[1][2] + M[1][3];
    TZ = X * M[2][0] + Y * M[2][1] + Z * M[2][2] + M[2][3];
    X = TX;
    Y = TY;
    Z = TZ;
}

void            LowFpuPrecision()
{
    long            memvar;
    _asm {
    finit		// initialize the FPU
    fwait		// wait for operation to complete
    fstcw[memvar]	// store FPU control word to memvar
    fwait		// wait for operation to complete
    mov eax,[memvar]	// move memvar to a register
    and eax, 0fffffcffh	// mask off precision bits to set to 24-bit precision
    mov[memvar], eax	// save control word to memory
    fldcw[memvar]	// load control word back to FPU
    fwait		// wait for operation to complete
    }
}

void           ChangeRGBs(int count)
{
    int             j;
    for (j = 0; j < count; j++) {
	red += 55;
	if (red > 255)
	    red -= 256;
	green += 10;
	if (green > 255)
	    green -= 256;
	blue += 20;
	if (blue > 255)
	    blue -= 256;
    }
}

void            RenderFace(MAT3D & M, float Xp, float Yp)
{
    //  Project four corners
    // upper left:              // (0,      0,      0) (in world space)
    vtx1.x = 0;
    vtx1.y = 0;
    vtx1.z = 0;
    // upper right:             // (height, 0,      0)
    vtx2.x = Xp;
    vtx2.y = 0;
    vtx2.z = 0;
    // lower left:              // (height, length, 0)
    vtx3.x = Xp;
    vtx3.y = Yp;
    vtx3.z = 0;
    // lower right:             // (0,      length, 0)
    vtx4.x = 0;
    vtx4.y = Yp;
    vtx4.z = 0;

    Translate_Point(M, vtx1.x, vtx1.y, vtx1.z);
    Translate_Point(M, vtx2.x, vtx2.y, vtx2.z);
    Translate_Point(M, vtx3.x, vtx3.y, vtx3.z);
    Translate_Point(M, vtx4.x, vtx4.y, vtx4.z);

    vtx1.oow = 1. / vtx1.z;
    vtx2.oow = 1. / vtx2.z;
    vtx3.oow = 1. / vtx3.z;
    vtx4.oow = 1. / vtx4.z;

    vtx1.x = vtx1.x * vtx1.oow + midx;	// x1 (in screen space)
    vtx1.y = vtx1.y * vtx1.oow + midy;	// y1

    vtx2.x = vtx2.x * vtx2.oow + midx;	// x2
    vtx2.y = vtx2.y * vtx2.oow + midy;	// y2

    vtx3.x = vtx3.x * vtx3.oow + midx;	// x3
    vtx3.y = vtx3.y * vtx3.oow + midy;	// y3

    vtx4.x = vtx4.x * vtx4.oow + midx;	// x4
    vtx4.y = vtx4.y * vtx4.oow + midy;	// y4

    SNAP(vtx1);
    SNAP(vtx2);
    SNAP(vtx3);
    SNAP(vtx4);

    // iterated color
    if (iterated) {
	vtx1.r = red;
	vtx1.g = green;
	vtx1.b = blue;
	ChangeRGBs(1);
	vtx2.r = red;
	vtx2.g = green;
	vtx2.b = blue;
	ChangeRGBs(1);
	vtx3.r = red;
	vtx3.g = green;
	vtx3.b = blue;
	ChangeRGBs(1);
	vtx4.r = red;
	vtx4.g = green;
	vtx4.b = blue;
	ChangeRGBs(1);
    } else
	// or flat color:
	grConstantColorValue(grWireColor);

/*
   Draw the 2 triangles:
   1 2
    /
   4 3
	*/

    guDrawTriangleWithClip(&vtx1, &vtx2, &vtx4);
    guDrawTriangleWithClip(&vtx2, &vtx4, &vtx3);
}

int             IsVisible(MAT3D & M)
{
    //  Point1 = M *    [0, 0, 0, 1]t
    float           X1 = M[0][3];
    float           Y1 = M[1][3];
    float           Z1 = M[2][3];

    //  Point2 = M *    [0, 1, 0, 1]t
    float           X2 = M[0][1] + M[0][3];
    float           Y2 = M[1][1] + M[1][3];
    float           Z2 = M[2][1] + M[2][3];

    //  Point3 = M *    [1, 0, 0, 1]t
    float           X3 = M[0][0] + M[0][3];
    float           Y3 = M[1][0] + M[1][3];
    float           Z3 = M[2][0] + M[2][3];

    float           D =
    -X1 * (Y2 * Z3 - Y3 * Z2)
    - X2 * (Y3 * Z1 - Y1 * Z3)
    - X3 * (Y1 * Z2 - Y2 * Z1);

    if (!inside)
	return D > 0;		// external view
    return D < 0;		// inside view
}

void            quit(char *t)
{
    if (GlideIsRunning)
	grGlideShutdown();
    GlideIsRunning = 0;
    if (*t) {
	printf("\n=== Error: %s!\n", t);
	exit(1);
    }
    puts("\nBye.");
    exit(0);
}

void		BuildCube(MAT3D & S1, MAT3D & S2,
			  MAT3D & S3, MAT3D & S4,
			  MAT3D & S5, MAT3D & S6)
{
    MAT3D           Flip, Temp;

    Flip = RotateY(180. * 3.141592654 / 180.);
    Temp = Translate(side1, 0, 0);
    Flip = Temp * Flip;

    S1 = Translate(0, 0, 0);

    Temp = RotateX(-90. * 3.141592654 / 180.);
    S2 = Temp * Flip;

    Temp = Translate(side2 - side1, 0, 0);
    S3 = Temp * Flip;
    Temp = RotateY(90. * 3.141592654 / 180.);
    S3 = Temp * S3;

    Temp = Translate(0, 0, side2);
    S4 = Temp * Flip;

    S5 = RotateX(-90. * 3.141592654 / 180.);
    Temp = Translate(0, side2, 0);
    S5 = Temp * S5;

    S6 = RotateY(90. * 3.141592654 / 180.);
    Temp = Translate(side1, 0, 0);
    S6 = Temp * S6;
}

void            startglide()
{
    grGlideInit();
    assert(grSstQueryHardware(&hwconfig));
    grSstSelect(0);
    assert(grSstWinOpen(0,
			GR_RESOLUTION_640x480,
			GR_REFRESH_60Hz,
			GR_COLORFORMAT_ABGR,
			GR_ORIGIN_UPPER_LEFT,
			2,
			1
			));

    GlideIsRunning = YES;

    if (iterated == NO)
    /* flat shading */
	grColorCombine( GR_COMBINE_FUNCTION_LOCAL,
		GR_COMBINE_FACTOR_NONE,
		GR_COMBINE_LOCAL_CONSTANT,
		GR_COMBINE_OTHER_NONE,
		FXFALSE );
    else
    /* iterated colors */
	grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
		GR_COMBINE_FACTOR_NONE,
		GR_COMBINE_LOCAL_ITERATED,
		GR_COMBINE_OTHER_NONE,
		FXFALSE);

    grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
    grDepthMask(FXTRUE);
}

void            init()
{
    startglide();

    LowFpuPrecision();

    midx = size_x / 2;
    midy = size_y / 2;
}

void            info()
{
#ifndef BANSHEE
    char            nfo[132];

    strcpy(nfo, "Mode: ");
    if (inside)
	strcat(nfo, "inside [v]iew, ");
    else
	strcat(nfo, "outside [v]iew, ");
    if (twosided)
	strcat(nfo, "two [s]ided, ");
    else
	strcat(nfo, "one [s]ided, ");
    if (iterated)
	strcat(nfo, "iterated [c]olors, ");
    else
	strcat(nfo, "flat [c]olors, ");
    if (allsides)
	strcat(nfo, "all [f]aces visible");
    else
	strcat(nfo, "some [f]aces visible");

    printf("%s\n", nfo);
#endif
}

void            main()
{
    int             j, k, done = NO;
    MAT3D           S[7], Side[7];
    int	     	    running = 1;

    puts("\n=== Spinning Cube Test ===");

    init();
    BuildCube(S[1], S[2], S[3], S[4], S[5], S[6]);
    info();

    while (!done) {
	if (running) {
	j++;
	MAT3D           Cube, Temp;

	// Spin it
	Cube = Translate(-128, -100, -100);
	Temp = RotateXYZ(j * 0.01, j * 0.001, j * 0.007);
	Cube = Temp * Cube;
	Temp = Translate(0, 0, sin(-j * 0.0121) * 350 + 600);
	Cube = Temp * Cube;
	Temp = ScaleXYZ(200, 200, 1);
	Cube = Temp * Cube;

	for (k = 1; k < 7; k++)
	    Side[k] = Cube * S[k];
	}

	red = 20;
	green = 10;
	blue = 200;

	grBufferClear(0x0, 0, GR_WDEPTHVALUE_FARTHEST);

	grWireColor = 0xaa0101;	// flat color
	if (allsides && (twosided || IsVisible(Side[1])))
	    RenderFace(Side[1], side1, side2);
	else
	    ChangeRGBs(4);

	grWireColor = 0xaa01aa;
	if ((twosided || IsVisible(Side[2])))
	    RenderFace(Side[2], side1, side2);
	else
	    ChangeRGBs(4);

	grWireColor = 0xaaaa01;
	if ((twosided || IsVisible(Side[3])))
	    RenderFace(Side[3], side2, side2);
	else
	    ChangeRGBs(4);

	grWireColor = 0x090909;
	if (allsides && (twosided || IsVisible(Side[4])))
	    RenderFace(Side[4], side1, side2);
	else
	    ChangeRGBs(4);

	grWireColor = 0x01aa01;
	if ((twosided || IsVisible(Side[5])))
	    RenderFace(Side[5], side1, side2);
	else
	    ChangeRGBs(4);

	grWireColor = 0x0101aa;
	if ((twosided || IsVisible(Side[6])))
	    RenderFace(Side[6], side2, side2);
	else
	    ChangeRGBs(4);

	grBufferSwap(1);

	if (kbhit())
	    switch (getch()) {
	    case 0x1b:		// VK_ESCAPE:
		done = YES;
		break;
	    case 0xD:		// VK_RETURN:
	    case 's':
		twosided = !twosided;
		info();
		break;
	    case 0x8:		// VK_BACK:
	    case 'v':
		inside = !inside;
		info();
		break;
	    case ' ':		// VK_SPACE:
		running = !running;
		break;
	    case 'c':
		iterated = !iterated;
		if (!iterated)
		    /* flat shading */
		    grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
				   GR_COMBINE_FACTOR_NONE,
				   GR_COMBINE_LOCAL_CONSTANT,
				   GR_COMBINE_OTHER_NONE,
				   FXFALSE);
		else
		    /* iterated colors */
		    grColorCombine(GR_COMBINE_FUNCTION_LOCAL,
				   GR_COMBINE_FACTOR_NONE,
				   GR_COMBINE_LOCAL_ITERATED,
				   GR_COMBINE_OTHER_NONE,
				   FXFALSE);
		info();
		break;
	    case '+':
	    case 'f':
		allsides = !allsides;
		info();
		break;
#ifndef BANSHEE
	    case 9:		// VK_TAB:
		twoscreens = !twoscreens;
		if (twoscreens)
		    grSstControl(GR_CONTROL_ACTIVATE);	 // 3dfx on
		else
		    grSstControl(GR_CONTROL_DEACTIVATE); // 3dfx off
		break;
#endif
	    }
	if (done)
	    break;
    }

    quit("");
}
